use std::{
    cell::{Cell, UnsafeCell},
    hint::cold_path,
    ptr::NonNull,
    sync::Arc,
    time::Duration,
};

use crossbeam_utils::CachePadded;

use crate::{
    context::{Context, Heartbeat},
    heartbeat::OwnedHeartbeatReceiver,
    job::{JobQueue as JobList, JobResult, QueuedJob as Job, QueuedJob, StackJob},
    latch::{AsCoreLatch, CoreLatch, Probe, WorkerLatch},
    util::DropGuard,
};

pub struct WorkerThread {
    pub(crate) context: Arc<Context>,
    pub(crate) queue: UnsafeCell<JobList>,
    pub(crate) heartbeat: OwnedHeartbeatReceiver,
    pub(crate) join_count: Cell<u8>,
}

thread_local! {
    static WORKER: UnsafeCell<Option<NonNull<WorkerThread>>> = const { UnsafeCell::new(None) };
}

impl WorkerThread {
    pub fn new_in(context: Arc<Context>) -> Self {
        let heartbeat = context.heartbeats.new_heartbeat();

        Self {
            context,
            queue: UnsafeCell::new(JobList::new()),
            heartbeat,
            join_count: Cell::new(0),
        }
    }
}

impl WorkerThread {
    pub fn run(self: Box<Self>) {
        let this = Box::into_raw(self);
        unsafe {
            Self::set_current(this);
        }

        let _guard = DropGuard::new(|| unsafe {
            // SAFETY: this is only called when the thread is exiting
            Self::unset_current();
            Self::drop_in_place(this);
        });

        tracing::trace!("WorkerThread::run: starting worker thread");

        unsafe {
            (&*this).run_inner();
        }

        tracing::trace!("WorkerThread::run: worker thread finished");
    }

    fn run_inner(&self) {
        let mut job = None;
        'outer: loop {
            if let Some(job) = job {
                self.execute(job);
            }

            if self.context.should_exit() {
                // if the context is stopped, break out of the outer loop which
                // will exit the thread.
                break 'outer;
            }

            // no more jobs, wait to be notified of a new job or a heartbeat.
            job = self.find_work_or_wait();
        }
    }
}

impl WorkerThread {
    pub(crate) fn find_work(&self) -> Option<NonNull<Job>> {
        self.find_work_inner().left()
    }

    /// Looks for work in the local queue, then in the shared context, and if no
    /// work is found, waits for the thread to be notified of a new job, after
    /// which it returns `None`.
    /// The caller should then check for `should_exit` to determine if the
    /// thread should exit, or look for work again.
    pub(crate) fn find_work_or_wait(&self) -> Option<NonNull<Job>> {
        match self.find_work_inner() {
            either::Either::Left(job) => {
                return Some(job);
            }
            either::Either::Right(mut guard) => {
                // no jobs found, wait for a heartbeat or a new job
                tracing::trace!("WorkerThread::find_work_or_wait: waiting for new job");
                self.heartbeat.latch().wait_with_lock(&mut guard);
                tracing::trace!("WorkerThread::find_work_or_wait: woken up from wait");
                None
            }
        }
    }

    #[inline]
    fn find_work_inner(
        &self,
    ) -> either::Either<NonNull<Job>, parking_lot::MutexGuard<'_, crate::context::Shared>> {
        // first check the local queue for jobs
        if let Some(job) = self.pop_front() {
            tracing::trace!("WorkerThread::find_work_inner: found local job: {:?}", job);
            return either::Either::Left(job);
        }

        // then check the shared context for jobs
        let mut guard = self.context.shared();

        if let Some(job) = guard.pop_job() {
            tracing::trace!("WorkerThread::find_work_inner: found shared job: {:?}", job);
            return either::Either::Left(job);
        }

        either::Either::Right(guard)
    }

    #[inline(always)]
    pub(crate) fn tick(&self) {
        if self.heartbeat.take() {
            tracing::trace!(
                "received heartbeat, thread id: {:?}",
                self.heartbeat.index()
            );
            self.heartbeat_cold();
        }
    }

    #[inline]
    fn execute(&self, job: NonNull<Job>) {
        self.tick();
        unsafe { Job::execute(job.as_ptr()) };
    }

    #[cold]
    fn heartbeat_cold(&self) {
        let mut guard = self.context.shared();

        if !guard.jobs.contains_key(&self.heartbeat.id()) {
            if let Some(job) = self.pop_back() {
                Job::set_shared(unsafe { job.as_ref() });
                tracing::trace!("heartbeat: sharing job: {:?}", job);
                guard.jobs.insert(self.heartbeat.id(), job);
                unsafe {
                    // SAFETY: we are holding the lock on the shared context.
                    self.context.notify_job_shared();
                }
            }
        }
    }
}

impl WorkerThread {
    #[inline]
    pub fn pop_back(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_back() }
    }

    #[inline]
    pub fn push_back(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_back(job) }
    }

    #[inline]
    pub fn pop_front(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_front() }
    }

    #[inline]
    pub fn push_front(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_front(job) }
    }
}

impl WorkerThread {
    #[inline]
    pub fn current_ref<'a>() -> Option<&'a Self> {
        unsafe { (*WORKER.with(UnsafeCell::get)).map(|ptr| ptr.as_ref()) }
    }

    unsafe fn set_current(this: *const Self) {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                assert!(
                    (&mut *cell.get())
                        .replace(NonNull::new_unchecked(
                            this as *const WorkerThread as *mut WorkerThread,
                        ))
                        .is_none()
                );
            }
        });
    }

    unsafe fn unset_current() {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                (&mut *cell.get()).take();
            }
        });
    }

    unsafe fn drop_in_place(this: *mut Self) {
        unsafe {
            // SAFETY: this is only called when the thread is exiting, so we can
            // safely drop the thread. We use `drop_in_place` to prevent `Box`
            // from creating a no-alias reference to the worker thread.
            core::ptr::drop_in_place(this);
            _ = Box::<core::mem::ManuallyDrop<Self>>::from_raw(this as _);
        }
    }
}

pub struct HeartbeatThread {
    ctx: Arc<Context>,
}

impl HeartbeatThread {
    const HEARTBEAT_INTERVAL: Duration = Duration::from_micros(100);

    pub fn new(ctx: Arc<Context>) -> Self {
        Self { ctx }
    }

    pub fn run(self) {
        tracing::trace!("new heartbeat thread {:?}", std::thread::current());

        let mut i = 0;
        loop {
            let sleep_for = {
                if self.ctx.should_exit() {
                    break;
                }

                self.ctx.heartbeats.notify_nth(i);
                let num_heartbeats = self.ctx.heartbeats.len();

                if i >= num_heartbeats {
                    i = 0;
                } else {
                    i += 1;
                }

                Self::HEARTBEAT_INTERVAL.checked_div(num_heartbeats as u32)
            };

            if let Some(duration) = sleep_for {
                std::thread::sleep(duration);
            }
        }
    }
}

impl WorkerThread {
    pub fn wait_until_queued_job<T>(
        &self,
        job: *const QueuedJob,
    ) -> Option<std::thread::Result<T>> {
        let recv = unsafe { (*job).as_receiver::<T>() };
        // we've already checked that the job was popped from the queue
        // check if shared job is our job
        if let Some(shared_job) = self.context.shared().jobs.remove(&self.heartbeat.id()) {
            if core::ptr::eq(shared_job.as_ptr(), job as *const Job as _) {
                // this is the job we are looking for, so we want to
                // short-circuit and call it inline
                return None;
            } else {
                // this isn't the job we are looking for, but we still need to
                // execute it
                unsafe { Job::execute(shared_job.as_ptr()) };
            }
        }

        // do the usual thing and wait for the job's latch
        loop {
            match recv.poll() {
                Some(t) => {
                    return Some(t);
                }
                None => {
                    cold_path();

                    // check local jobs before locking shared context
                    if let Some(job) = self.find_work_or_wait() {
                        tracing::trace!(
                            "thread {:?} executing local job: {:?}",
                            self.heartbeat.index(),
                            job
                        );
                        unsafe {
                            Job::execute(job.as_ptr());
                        }
                        tracing::trace!(
                            "thread {:?} finished local job: {:?}",
                            self.heartbeat.index(),
                            job
                        );
                        continue;
                    }
                }
            }
        }
    }

    pub fn wait_until_latch<L>(&self, latch: &L)
    where
        L: Probe,
    {
        if !latch.probe() {
            tracing::trace!("thread {:?} waiting on latch", self.heartbeat.index());
            self.wait_until_latch_cold(latch);
        }
    }

    #[cold]
    fn wait_until_latch_cold<L>(&self, latch: &L)
    where
        L: Probe,
    {
        if let Some(shared_job) = self.context.shared().jobs.remove(&self.heartbeat.id()) {
            tracing::trace!(
                "thread {:?} reclaiming shared job: {:?}",
                self.heartbeat.index(),
                shared_job
            );
            unsafe { Job::execute(shared_job.as_ptr()) };
        }

        // do the usual thing and wait for the job's latch
        // do the usual thing??? chatgipity really said this..
        while !latch.probe() {
            // check local jobs before locking shared context
            if let Some(job) = self.find_work_or_wait() {
                tracing::trace!(
                    "thread {:?} executing local job: {:?}",
                    self.heartbeat.index(),
                    job
                );
                unsafe {
                    Job::execute(job.as_ptr());
                }
                tracing::trace!(
                    "thread {:?} finished local job: {:?}",
                    self.heartbeat.index(),
                    job
                );
                continue;
            }
        }
    }
}
